using Kafka.Public;
using Kafka.Public.Loggers;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;

namespace franz
{
    public class Telefranz
    {
        static class _TelefranzConsumers<T> where T : silver_messages.message, new() {
            internal static readonly Dictionary<Action<T>, Action<KafkaRecord<string, string>>> wrappings 
                = new Dictionary<Action<T>, Action<KafkaRecord<string, string>>>();
        }
        private string handling_group { get; set; } = "Liszt";
        private silver_messages.global.report howToReport { get; set; }
        private ClusterClient clusterClient;
        private Dictionary<string, KafkaConsumer<string, string>> topicConsumers {get;set;}
             = new Dictionary<string, KafkaConsumer<string, string>>();
        private Dictionary<string, int> topicSubscribers {get;set;} = new Dictionary<string, int>();
        private Kafka.Public.Configuration clusterClientConfig = null;
        private static ConsoleLogger consoleLogger = new ConsoleLogger();

        public Telefranz(string name, string bootstrap_servers,
            List<string> commands = null, List<string> checks = null,
            List<string> errors = null, List<string> warnings = null)
        {

            var serializer = new StringSerializer();
            var deserializer = new StringDeserializer();
            var serializationConfig = new SerializationConfig(){SerializeOnProduce = true};
            serializationConfig.SetDefaultSerializers(serializer, serializer);
            serializationConfig.SetDefaultDeserializers(deserializer, deserializer);
            clusterClientConfig = new Kafka.Public.Configuration { 
                Seeds = bootstrap_servers, 
                ClientId = name,
                SerializationConfig = serializationConfig
            };
            clusterClient = new ClusterClient(clusterClientConfig, consoleLogger);

            handling_group = name;
            this.howToReport = new silver_messages.global.report()
            {
                name = name,
                errors = errors ?? new List<string>(),
                warnings = warnings ?? new List<string>()
            };
            if (commands?.Count > 0)
            {
                howToReport.capabilites.commands = commands;
            }
            if (checks?.Count > 0)
            {
                howToReport.capabilites.checks = checks;
            }

            addHandler<silver_messages.global.sound_off>((m) =>
            {
                ProduceMessage(this.howToReport);
            });
            addHandler<silver_messages.global.stop>((m) =>
            {
                if (m.name == handling_group)
                {
                    Environment.Exit(0);
                }
            });
            addHandler<silver_messages.global.restart>((m) =>
            {
                if (m.name == handling_group)
                {
                    throw new NotImplementedException();//TODO
                }
            });
        }

        public void addHandler<T>(Action<T> theAction) where T : silver_messages.message, new()
        {
            var dummy = new T();
            Console.WriteLine($"created dummy, it reports as {dummy}. typeof(T) reports as {typeof(T)}. Topic is {dummy.topic}");
            var topic = dummy.topic;
            
            Action<KafkaRecord<string, string>> wrapped = (kr) =>
            {
                Console.WriteLine($"wrapped receives kafkarecord, is ready to process: {kr.Value}");
                var deserialized = JsonConvert.DeserializeObject<T>(kr.Value);
                if(deserialized != null && deserialized.type == deserialized.GetType().ToString())
                {
                    Console.WriteLine("message deserialized for this handler");
                    theAction(deserialized);
                    Console.WriteLine("action survived");
                }
                else
                {
                    Console.WriteLine("message didn't deserialize for this handler");
                }
            };
            if(!topicConsumers.ContainsKey(topic)){
                Console.WriteLine($"making consumer for {handling_group}, topic {topic}");
                topicConsumers[topic] = new KafkaConsumer<string, string>(topic, clusterClient);
                topicConsumers[topic].ConsumeFromLatest();
                topicSubscribers.Add(topic, 0);
            }
            topicConsumers[topic].MessageReceived += wrapped;
            _TelefranzConsumers<T>.wrappings[theAction] = wrapped;
            topicSubscribers[topic]++;
        }
        public void removeHandler<T>(Action<T> theAction) where T : silver_messages.message, new()
        {
            var dummy = new T();
            topicConsumers[dummy.topic].MessageReceived -= _TelefranzConsumers<T>.wrappings[theAction];
            _TelefranzConsumers<T>.wrappings.Remove(theAction);
            topicSubscribers[dummy.topic]--;
            if(topicSubscribers[dummy.topic] == 0)
            {
                topicConsumers[dummy.topic].StopConsume();
            }
        }

        public void ProduceMessage<T>(T message) where T : silver_messages.message
        {
            Console.WriteLine($"producing message {message}");
            clusterClient.Produce(message.topic, message.ToString());
        }
    }
}